@@ -3,6 +3,7 @@ import pickle
 import logging
 import os
 import hashlib
+from shutil import move, rmtree
 import tempfile
 import inspect
 from fsspec import AbstractFileSystem, filesystem
@@ -112,6 +113,9 @@ class CachingFileSystem(AbstractFileSystem):
         self.load_cache()
         self.fs = fs if fs is not None else filesystem(target_protocol, **self.kwargs)
 
+    def _mkcache(self):
+        os.makedirs(self.storage[-1], exist_ok=True)
+
     def load_cache(self):
         """Read set of stored blocks from file"""
         cached_files = []
@@ -120,10 +124,14 @@ class CachingFileSystem(AbstractFileSystem):
             if os.path.exists(fn):
                 with open(fn, "rb") as f:
                     # TODO: consolidate blocks here
-                    cached_files.append(pickle.load(f))
+                    loaded_cached_files = pickle.load(f)
+                    for c in loaded_cached_files.values():
+                        if isinstance(c["blocks"], list):
+                            c["blocks"] = set(c["blocks"])
+                    cached_files.append(loaded_cached_files)
             else:
-                os.makedirs(storage, exist_ok=True)
                 cached_files.append({})
+        self._mkcache()
         self.cached_files = cached_files or [{}]
         self.last_cache = time.time()
 
@@ -156,10 +164,12 @@ class CachingFileSystem(AbstractFileSystem):
         fn2 = tempfile.mktemp()
         with open(fn2, "wb") as f:
             pickle.dump(cache, f)
-        os.replace(fn2, fn)
+        self._mkcache()
+        move(fn2, fn)
 
     def _check_cache(self):
         """Reload caches if time elapsed or any disappeared"""
+        self._mkcache()
         if not self.cache_check:
             # explicitly told not to bother checking
             return
@@ -170,14 +180,18 @@ class CachingFileSystem(AbstractFileSystem):
 
     def _check_file(self, path):
         """Is path in cache and still valid"""
+        path = self._strip_protocol(path)
+
         self._check_cache()
         if not path.startswith(self.target_protocol):
             store_path = self.target_protocol + "://" + path
             path = self.fs._strip_protocol(store_path)
+        else:
+            store_path = path
         for storage, cache in zip(self.storage, self.cached_files):
-            if path not in cache:
+            if store_path not in cache:
                 continue
-            detail = cache[path].copy()
+            detail = cache[store_path].copy()
             if self.check_files:
                 if detail["uid"] != self.fs.ukey(path):
                     continue
@@ -189,6 +203,41 @@ class CachingFileSystem(AbstractFileSystem):
                 return detail, fn
         return False, None
 
+    def clear_cache(self):
+        """Remove all files and metadat from the cache
+
+        In the case of multiple cache locations, this clears only the last one,
+        which is assumed to be the read/write one.
+        """
+        rmtree(self.storage[-1])
+        self.load_cache()
+
+    def pop_from_cache(self, path):
+        """Remove cached version of given file
+
+        Deletes local copy of the given (remote) path. If it is found in a cache
+        location which is not the last, it is assumed to be read-only, and
+        raises PermissionError
+        """
+        path = self._strip_protocol(path)
+        if not path.startswith(self.target_protocol):
+            store_path = self.target_protocol + "://" + path
+            path = self.fs._strip_protocol(store_path)
+        else:
+            store_path = path
+        _, fn = self._check_file(path)
+        if fn is None:
+            return
+        if fn.startswith(self.storage[-1]):
+            # is in in writable cache
+            os.remove(fn)
+            self.cached_files[-1].pop(store_path)
+            self.save_cache()
+        else:
+            raise PermissionError(
+                "Can only delete cached file in last, writable cache location"
+            )
+
     def _open(
         self,
         path,
@@ -250,6 +299,7 @@ class CachingFileSystem(AbstractFileSystem):
             logger.debug("Creating local sparse file for %s" % path)
 
         # call target filesystems open
+        self._mkcache()
         f = self.fs._open(
             path,
             mode=mode,
@@ -304,12 +354,17 @@ class CachingFileSystem(AbstractFileSystem):
             "__reduce__",
             "open",
             "cat",
+            "cat_file",
             "get",
             "read_block",
             "tail",
             "head",
             "_check_file",
             "_check_cache",
+            "clear_cache",
+            "pop_from_cache",
+            "_mkcache",
+            "local_file",
         ]:
             # all the methods defined in this class. Note `open` here, since
             # it calls `_open`, but is actually in superclass
@@ -361,6 +416,7 @@ class WholeFileCacheFileSystem(CachingFileSystem):
     """
 
     protocol = "filecache"
+    local_file = True
 
     def _open(self, path, mode="rb", **kwargs):
         path = self._strip_protocol(path)
@@ -407,6 +463,7 @@ class WholeFileCacheFileSystem(CachingFileSystem):
                 else self.compression
             )
             f = compr[comp](f, mode="rb")
+        self._mkcache()
         with open(fn, "wb") as f2:
             if isinstance(f, AbstractBufferedFile):
                 # want no type of caching if just downloading whole thing
@@ -433,9 +490,15 @@ class SimpleCacheFileSystem(CachingFileSystem):
     does not keep any metadata about the download time or file details.
     It is therefore safer to use in multi-threaded/concurrent situations.
 
+    This is the only of the caching filesystems that supports write: you will
+    be given a real local open file, and upon close and commit, it will be
+    uploaded to the target filesystem; the writability or the target URL is
+    not checked until that time.
+
     """
 
     protocol = "simplecache"
+    local_file = True
 
     def __init__(self, **kwargs):
         kw = kwargs.copy()
@@ -448,6 +511,7 @@ class SimpleCacheFileSystem(CachingFileSystem):
         self.cached_files = [{}]
 
     def _check_file(self, path):
+        self._check_cache()
         sha = hash_name(path, self.same_names)
         for storage in self.storage:
             fn = os.path.join(storage, sha)
@@ -469,7 +533,7 @@ class SimpleCacheFileSystem(CachingFileSystem):
             store_path = path
         path = self.fs._strip_protocol(store_path)
         if "r" not in mode:
-            return self.fs._open(path, mode=mode, **kwargs)
+            return LocalTempFile(self, path, mode=mode)
         fn = self._check_file(path)
         if fn:
             return open(fn, mode)
@@ -479,6 +543,7 @@ class SimpleCacheFileSystem(CachingFileSystem):
         logger.debug("Copying %s to local cache" % path)
         kwargs["mode"] = mode
 
+        self._mkcache()
         with self.fs._open(path, **kwargs) as f, open(fn, "wb") as f2:
             if isinstance(f, AbstractBufferedFile):
                 # want no type of caching if just downloading whole thing
@@ -509,6 +574,51 @@ class SimpleCacheFileSystem(CachingFileSystem):
         return self._open(path, mode)
 
 
+class LocalTempFile:
+    """A temporary local file, which will be uploaded on commit"""
+
+    def __init__(self, fs, path, fn=None, mode="wb", autocommit=True, seek=0):
+        fn = fn or tempfile.mktemp()
+        self.mode = mode
+        self.fn = fn
+        self.fh = open(fn, mode)
+        if seek:
+            self.fh.seek(seek)
+        self.path = path
+        self.fs = fs
+        self.closed = False
+        self.autocommit = autocommit
+
+    def __reduce__(self):
+        # always open in rb+ to allow continuing writing at a location
+        return (
+            LocalTempFile,
+            (self.fs, self.path, self.fn, "rb+", self.autocommit, self.tell()),
+        )
+
+    def __enter__(self):
+        return self.fh
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        self.close()
+
+    def close(self):
+        self.fh.close()
+        self.closed = True
+        if self.autocommit:
+            self.commit()
+
+    def discard(self):
+        self.fh.close()
+        os.remove(self.fn)
+
+    def commit(self):
+        self.fs.put(self.fn, self.path)
+
+    def __getattr__(self, item):
+        return getattr(self.fh, item)
+
+
 def hash_name(path, same_name):
     if same_name:
         hash = os.path.basename(path)
